@@ -1,6 +1,7 @@
-from .data import COINNDataLoader, COINNDataset
+from .data import COINNDataset, COINNDataHandle
 from .distrib import COINNLocal, COINNRemote
 from .trainer import COINNTrainer
 from .nn import PooledTrainer
 from .distrib import COINNLearner, COINNReducer
+from.io import COINPyService
 
