@@ -53,16 +53,3 @@ def save_grads(file_path, grads):
 
 def load_grads(file_path):
     return _np.load(file_path, allow_pickle=True)
-
-
-def get_safe_batch_size(batch_size, dataset_len):
-    if dataset_len % batch_size == 0:
-        return batch_size
-
-    for i, j in zip(range(batch_size, 1, -1), range(batch_size, batch_size * 2, 1)):
-        if dataset_len % i != 1:
-            return i
-        if dataset_len % j != 1:
-            return j
-
-    return batch_size
